package at.ac.uibk.gitsearch.service;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import at.ac.uibk.gitsearch.domain.util.TreeNode;
import java.util.Optional;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO.PublicVisibility;
import org.junit.jupiter.api.Test;

public class ZipRepackagingServiceTest {

    protected static class MockedSingletonTreeNodeForFiltering extends TreeNode<SearchResultDTO> {

        private String[] filterStrings;

        /**
         *
         * @param path
         * @param separator
         *            just a technical parameter, in order to separate path from
         *            filters
         * @param filterStrings
         */
        public MockedSingletonTreeNodeForFiltering(String path, Boolean separator, String... filterStrings) {
            this(filterStrings);
            this.getData().getFile().setPath(path);
        }

        public MockedSingletonTreeNodeForFiltering(String... filterStrings) {
            super(
                new SearchResultDTO() {
                    MetadataFile file = new MetadataFile();

                    UserProvidedMetadataDTO userMetaData = new UserProvidedMetadataDTO();

                    {
                        PublicVisibility pv = new PublicVisibility();
                        pv.setExcept(filterStrings);
                        userMetaData.setPublicVisibility(pv);
                    }

                    public MetadataFile getFile() {
                        return file;
                    }

                    public UserProvidedMetadataDTO getMetadata() {
                        return userMetaData;
                    }
                }
            );
            this.filterStrings = filterStrings;
            this.getData().getFile().setPath("metadata.yaml");
        }

        public String[] getFilterStrings() {
            return filterStrings;
        }
    }

    private Optional<TreeNode<SearchResultDTO>> getMockedFilterTreeNode(String... filters) {
        return Optional.of(new MockedSingletonTreeNodeForFiltering(filters));
    }

    @Test
    public void simpleFilterOutTests() {
        assertFalse(ZipRepackagingService.filterOut("someFile.txt", getMockedFilterTreeNode(), false));
        assertFalse(ZipRepackagingService.filterOut("someFile.txt", getMockedFilterTreeNode("some"), false));
        assertTrue(ZipRepackagingService.filterOut("someFile.txt", getMockedFilterTreeNode("some*"), false));
        assertFalse(ZipRepackagingService.filterOut("someFile.txt", getMockedFilterTreeNode("some*.pdf"), false));
        assertTrue(ZipRepackagingService.filterOut("someFile.txt", getMockedFilterTreeNode("some*.txt"), false));
    }

    @Test
    public void filterOutTestsWithLeadingRoot() {
        assertFalse(ZipRepackagingService.filterOut("/someFile.txt", getMockedFilterTreeNode(), false));
        assertFalse(ZipRepackagingService.filterOut("/someFile.txt", getMockedFilterTreeNode("**/some"), false));
        assertTrue(ZipRepackagingService.filterOut("/someFile.txt", getMockedFilterTreeNode("**/some*"), false));
        assertFalse(ZipRepackagingService.filterOut("/someFile.txt", getMockedFilterTreeNode("**/some*.pdf"), false));
        assertTrue(ZipRepackagingService.filterOut("/someFile.txt", getMockedFilterTreeNode("**/some*.txt"), false));

        assertTrue(ZipRepackagingService.filterOut("parent/someFile.txt", getMockedFilterTreeNode("**/some*.txt"), false));
        assertFalse(ZipRepackagingService.filterOut("parent/someFile.txt", getMockedFilterTreeNode("**/someFile.txt/"), false));
    }

    @Test
    public void filterOutTestsWithDirectories() {
        assertFalse(ZipRepackagingService.filterOut("someDir", getMockedFilterTreeNode(), true));
        assertFalse(ZipRepackagingService.filterOut("someDir", getMockedFilterTreeNode("some/"), true));
        assertTrue(ZipRepackagingService.filterOut("someDir", getMockedFilterTreeNode("/some*"), true));
        assertTrue(ZipRepackagingService.filterOut("someDir", getMockedFilterTreeNode("some*"), true));
        assertTrue(ZipRepackagingService.filterOut("someDir/", getMockedFilterTreeNode("some*/"), true));
        assertFalse(ZipRepackagingService.filterOut("someDir", getMockedFilterTreeNode("*/some/"), true));
        assertFalse(ZipRepackagingService.filterOut("parent/someDir", getMockedFilterTreeNode("*/some/"), true));
        assertTrue(ZipRepackagingService.filterOut("parent/someDir", getMockedFilterTreeNode("someDir/"), true));
        assertTrue(ZipRepackagingService.filterOut("parent/someDir", getMockedFilterTreeNode("*/someDir"), true));
        assertTrue(ZipRepackagingService.filterOut("parent/someDir", getMockedFilterTreeNode("*/some*"), true));
        assertTrue(ZipRepackagingService.filterOut("parent/someDir", getMockedFilterTreeNode("**/someDir"), true));
        assertTrue(ZipRepackagingService.filterOut("parent/someDir", getMockedFilterTreeNode("**/some*"), true));

        assertTrue(ZipRepackagingService.filterOut("parent/someDir", getMockedFilterTreeNode("parent"), true));
    }

    @Test
    public void filterOutTestsWithPathes() {
        assertFalse(ZipRepackagingService.filterOut("/root/parent/someDir", getMockedFilterTreeNode(), true));
        assertFalse(ZipRepackagingService.filterOut("/root/parent/someDir", getMockedFilterTreeNode("parent/some"), true));
        assertTrue(ZipRepackagingService.filterOut("/root/parent/someDir", getMockedFilterTreeNode("*/*rent/some*"), true));
        assertFalse(ZipRepackagingService.filterOut("/root/parent/someDir", getMockedFilterTreeNode("parent/some*"), true));
        assertFalse(ZipRepackagingService.filterOut("/root/parent/someDir", getMockedFilterTreeNode("par*/some*"), true));
        assertTrue(ZipRepackagingService.filterOut("/root/parent/someDir", getMockedFilterTreeNode("*/parent/some*"), true));
        assertTrue(ZipRepackagingService.filterOut("/root/parent/someDir", getMockedFilterTreeNode("*/*/some*"), true));
        assertTrue(ZipRepackagingService.filterOut("/root/parent/someDir", getMockedFilterTreeNode("**/some*"), true));
    }

    @Test
    public void complexFilterOutInHierarchy() {
        // complex szenario
        // Top Level: Forbid all "solutions" folders
        //       SubNode "1": Forbid all "test" and "Test" folders
        //          SubNode: "1/a": forbid all "excerise/*.txt" files
        //          SubNode: "1/b": allow all
        // SubNode "2": Forbid all "secret" folders
        MockedSingletonTreeNodeForFiltering root = new MockedSingletonTreeNodeForFiltering("", true, "solution/");
        MockedSingletonTreeNodeForFiltering node1 = new MockedSingletonTreeNodeForFiltering("1/metaData.yaml", true, "test/", "Test/");
        MockedSingletonTreeNodeForFiltering node1a = new MockedSingletonTreeNodeForFiltering("1/a/metaData.yaml", true, "exercise/*.txt");
        MockedSingletonTreeNodeForFiltering node1b = new MockedSingletonTreeNodeForFiltering("1/b/metaData.yaml", true);
        MockedSingletonTreeNodeForFiltering node2 = new MockedSingletonTreeNodeForFiltering(
            "2/metaData.yaml",
            true,
            "secret/",
            "secret*.txt"
        );

        root.addChild(node1);
        root.addChild(node2);
        node1.addChild(node1a);
        node1.addChild(node1b);

        Optional<TreeNode<SearchResultDTO>> rootO = Optional.of(root);
        Optional<TreeNode<SearchResultDTO>> node1O = Optional.of(node1);

        // restrictions from root
        assertFalse(ZipRepackagingService.filterOut("", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("something", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("something", rootO, false));
        assertTrue(ZipRepackagingService.filterOut("solution", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("solution", rootO, false));
        assertTrue(ZipRepackagingService.filterOut("something/solution", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("/something/solution", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("node1/solution", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("1/solution", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("1/solution", rootO, false));
        assertTrue(ZipRepackagingService.filterOut("2/solution", rootO, true));

        // restrictions from node 1
        assertFalse(ZipRepackagingService.filterOut("/something/test", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("/1/test", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("/1/Test", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("/1/Test/undMehr.txt", rootO, false));
        assertTrue(ZipRepackagingService.filterOut("/1/Test/undMehr", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("/1/Test", rootO, false));
        assertFalse(ZipRepackagingService.filterOut("/1/Testing", rootO, false));
        assertFalse(ZipRepackagingService.filterOut("/2/Test", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("/2/Testing", rootO, true));

        // restrictions from node 1/a
        assertFalse(ZipRepackagingService.filterOut("/exercise", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("/1/exercise", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("/1/a/exercise", rootO, true));

        assertTrue(ZipRepackagingService.filterOut("/1/a/exercise/something.txt", rootO, false));
        assertTrue(ZipRepackagingService.filterOut("/1/a/exercise/something2.txt", rootO, false));
        assertTrue(ZipRepackagingService.filterOut("/1/a/exercise/something2.txt", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("/1/a/exercise2/something2.txt", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("/1/a/exercise/something2.txt", rootO, true));

        // counterexamples on node 1/b for restrictions from node 1/a
        assertFalse(ZipRepackagingService.filterOut("/1/b/exercise/something.txt", rootO, false));
        assertFalse(ZipRepackagingService.filterOut("/1/b/exercise/something2.txt", rootO, false));
        assertFalse(ZipRepackagingService.filterOut("/1/b/exercise/something2.txt", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("/1/b/exercise/something2.txt", rootO, true));
        assertFalse(ZipRepackagingService.filterOut("/1/b/exercise/something.txt", rootO, true));

        // restrictions from node 2
        assertTrue(ZipRepackagingService.filterOut("/2/secret", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("/2/secret/tests", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("/2/someDir/secretFile.txt", rootO, true));
        assertTrue(ZipRepackagingService.filterOut("/2/secret/secretFile.txt", rootO, true));

        // testing on subNodes

        // restrictions from node 1
        assertFalse(ZipRepackagingService.filterOut("/something/test", node1O, true));
        assertTrue(ZipRepackagingService.filterOut("/1/test", node1O, true));
        assertTrue(ZipRepackagingService.filterOut("/1/Test", node1O, true));
        assertTrue(ZipRepackagingService.filterOut("/1/Test/undMehr.txt", node1O, false));
        assertTrue(ZipRepackagingService.filterOut("/1/Test/undMehr", node1O, true));
        assertFalse(ZipRepackagingService.filterOut("/1/Test", node1O, false));
        assertFalse(ZipRepackagingService.filterOut("/1/Testing", node1O, false));
        assertFalse(ZipRepackagingService.filterOut("/2/Test", node1O, true));
        assertFalse(ZipRepackagingService.filterOut("/2/Testing", node1O, true));

        // restrictions from node 1/a
        assertFalse(ZipRepackagingService.filterOut("/exercise", node1O, true));
        assertFalse(ZipRepackagingService.filterOut("/1/exercise", node1O, true));
        assertFalse(ZipRepackagingService.filterOut("/1/a/exercise", node1O, true));

        assertTrue(ZipRepackagingService.filterOut("/1/a/exercise/something.txt", node1O, false));
        assertTrue(ZipRepackagingService.filterOut("/1/a/exercise/something2.txt", node1O, false));
        assertTrue(ZipRepackagingService.filterOut("/1/a/exercise/something2.txt", node1O, true));
        assertFalse(ZipRepackagingService.filterOut("/1/a/exercise2/something2.txt", node1O, true));
        assertTrue(ZipRepackagingService.filterOut("/1/a/exercise/something2.txt", node1O, true));
    }
}
